/**
 * @file
 * @brief Implementation of rewriter protocols
 *
 * This module contains the algorithms for parsing, validating,
 * and encoding specific protocols.
 *
 * These protocol definitions are used by protocol.c.
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "redir.h"
#include "protocol_def.h"

/********************************************************
 **********  Helper Function Declarations ***************
 *******************************************************/

// Parse the given line, finding delimiter and returning number of fields
int protocol_explode_line (char *line, char **fields, size_t max_fields);

// return -1 if there's a request id in this field collection
int protocol_def_squid_reqid (char **fields, size_t field_cnt);

// Parse and validate for squid25 given these fields and field count
int protocol_def_squid25_vparse (protocol_input_line *input, char **fields, size_t field_cnt);

// Parse and validate for squid26 given these fields and field count
int protocol_def_squid26_vparse (protocol_input_line *input, char **fields, size_t field_cnt);

// Parse and validate for squid27 given these fields and field count
int protocol_def_squid27_vparse (protocol_input_line *input, char **fields, size_t field_cnt);

// Parse and validate for squid30 given these fields and field count
int protocol_def_squid30_vparse (protocol_input_line *input, char **fields, size_t field_cnt);

/********************************************************
 *****************  Helper Macros ***********************
 *******************************************************/

// This macro uses snprintf to write to a buffer, watching the length
// and then increments the buffer pointer and decrements the buffer length
// It also knows how to substitute a different value in case of NULL
// and it knows how to return ENOMEM
// It also automatically adds a space after the printed value
#define BUFFER_WRITE_MACRO(BUFPTR, BUFLEN, VALUE, DEFAULTVALUE)               \
  {                                                                           \
	  int slen;                                                               \
	  slen = snprintf (BUFPTR, BUFLEN, "%s ",                                 \
			           (VALUE) == NULL ? (DEFAULTVALUE) : (VALUE));           \
	  if (slen < 0 || slen >= (int) BUFLEN)                                   \
	  {                                                                       \
		  errno = ENOMEM;                                                     \
		  return -1;                                                          \
	  }                                                                       \
	  BUFPTR += slen;                                                         \
	  BUFLEN -= slen;                                                         \
  }

/********************************************************
 **************** Public Function Definitions ***********
 *******************************************************/

/**
 * @brief Parse the Squid 3.2 protocol
 *
 * @param input the line that must be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int squid32_parse (protocol_input_line *input)
{
	return squid30_parse (input);
}

/**
 * @brief Parse the Squid 3.1 protocol
 *
 * @param input the line that must be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int squid31_parse (protocol_input_line *input)
{
	return squid30_parse (input);
}

/**
 * @brief Parse the Squid 3.0 protocol
 *
 * @param input the line that must be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int squid30_parse (protocol_input_line *input)
{
#define MAX_SQUID30_FIELDS (1 + 4 + MAX_PROTOCOLS_OTHER_OPTIONS)
	// can have req id + 4 squid 26 fields + lots of options
	char *fields[MAX_SQUID30_FIELDS];
	int field_cnt;

	// Explode the line
	field_cnt = protocol_explode_line (input->line_buffer, fields, MAX_SQUID30_FIELDS);

	if (field_cnt < 4)
	{
		errno = EIO;
		return -1;
	}

	// Otherwise, parse and validate
	return protocol_def_squid30_vparse (input, fields, field_cnt);
}

/**
 * @brief Parse the Squid 2.7 protocol
 *
 * @param input the line that must be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int squid27_parse (protocol_input_line *input)
{
#define MAX_SQUID27_FIELDS (1 + 5 + MAX_PROTOCOLS_OTHER_OPTIONS)
	// can have req id + 5 squid 2.6 fields + lots of options
	char *fields[MAX_SQUID27_FIELDS];
	int field_cnt;

	// Explode the line
	field_cnt = protocol_explode_line (input->line_buffer, fields, MAX_SQUID27_FIELDS);

	if (field_cnt < 5)
	{
		errno = EIO;
		return -1;
	}

	// Otherwise, parse and validate
	return protocol_def_squid27_vparse (input, fields, field_cnt);
}

/**
 * @brief Parse the Squid 2.6 protocol
 *
 * @param input the line that must be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int squid26_parse (protocol_input_line *input)
{
	char *fields[6];
	int field_cnt;

	// Explode the line
	field_cnt = protocol_explode_line (input->line_buffer, fields, 6);

	if (field_cnt < 5)
	{
		errno = EIO;
		return -1;
	}

	// Otherwise, parse and validate
	return protocol_def_squid26_vparse (input, fields, field_cnt);
}

/**
 * @brief Parse the Squid 2.5 protocol
 *
 * @param input the line that must be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int squid25_parse (protocol_input_line *input)
{
	char *fields[4];

	// Explode the line
	if (protocol_explode_line (input->line_buffer, fields, 4) != 4)
	{
		errno = EIO;
		return -1;
	}

	// Otherwise, parse and validate
	return protocol_def_squid25_vparse (input, fields, 4);
}

/**
 * @brief Parse the Squid 2.4 protocol
 *
 * @param input the line that must be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int squid24_parse (protocol_input_line *input)
{
	return squid25_parse(input);
}

/**
 * @brief Encode a request in the Squid 2.4 protocol
 *
 * @param conf the current redirector configuration options
 * @param input the structure which holds the request data
 * @buffer the buffer in which the request should be created
 * @buffer_len the maximum number of bytes that can be put into the buffer
 *
 * @return 0 if the request was encoded successfully
 *        -1 if there was an error, and errno will be set:
 *           ENOMEM - There was a memory allocation error
 */
int squid24_encode (GlobalConf *conf, const protocol_input_line *input, char *buffer, size_t buffer_len)
{
	return squid25_encode(conf, input, buffer, buffer_len);
}

/**
 * @brief Encode a request in the Squid 2.5 protocol
 *
 * @param conf the current redirector configuration options
 * @param input the structure which holds the request data
 * @buffer the buffer in which the request should be created
 * @buffer_len the maximum number of bytes that can be put into the buffer
 *
 * @return 0 if the request was encoded successfully
 *        -1 if there was an error, and errno will be set:
 *           ENOMEM - There was a memory allocation error
 */
int squid25_encode (__attribute__ ((unused)) GlobalConf *conf, const protocol_input_line *input, char *buffer, size_t buffer_len)
{
	// Make sure we have all the data we need
	if (input->url == NULL)
	{
		errno = ENOENT;
		return -1;
	}

	// Start writing
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->url, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->IP, "-/-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->username, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->method, "-");

	// Back up one cuz that last space should be a CRLF
	buffer--;
	*buffer = '\n';

	return 0;
}

/**
 * @brief Encode a request in the Squid 2.6 protocol
 *
 * @param conf the current redirector configuration options
 * @param input the structure which holds the request data
 * @buffer the buffer in which the request should be created
 * @buffer_len the maximum number of bytes that can be put into the buffer
 *
 * @return 0 if the request was encoded successfully
 *        -1 if there was an error, and errno will be set:
 *           ENOMEM - There was a memory allocation error
 */
int squid26_encode (GlobalConf *conf, const protocol_input_line *input, char *buffer, size_t buffer_len)
{
	// Make sure we have all the data we need
	if (input->url == NULL)
	{
		errno = ENOENT;
		return -1;
	}

	// Start writing
	if(input->req_id != NULL && conf->server_concurrency) {
		BUFFER_WRITE_MACRO(buffer, buffer_len, input->req_id, "-");
	}
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->url, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->IP, "-/-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->username, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->method, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->group, "-");

	// Back up one cuz that last space should be a CRLF
	buffer--;
	*buffer = '\n';

	return 0;
}

/**
 * @brief Encode a request in the Squid 2.7 protocol
 *
 * @param conf the current redirector configuration options
 * @param input the structure which holds the request data
 * @buffer the buffer in which the request should be created
 * @buffer_len the maximum number of bytes that can be put into the buffer
 *
 * @return 0 if the request was encoded successfully
 *        -1 if there was an error, and errno will be set:
 *           ENOMEM - There was a memory allocation error
 */
int squid27_encode (GlobalConf *conf, const protocol_input_line *input, char *buffer, size_t buffer_len)
{
	int otheroption_num;

	// Make sure we have all the data we need
	if (input->url == NULL)
	{
		errno = ENOENT;
		return -1;
	}

	// Start writing
	if(input->req_id != NULL && conf->server_concurrency) {
		BUFFER_WRITE_MACRO(buffer, buffer_len, input->req_id, "-");
	}
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->url, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->IP, "-/-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->username, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->method, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->group, "-");

	// Now write our other options if we have them
	for (otheroption_num = 0;
		otheroption_num < MAX_PROTOCOLS_OTHER_OPTIONS &&
		  input->otheroptions[otheroption_num] != NULL;
		otheroption_num++)
	{
		BUFFER_WRITE_MACRO(buffer, buffer_len, input->otheroptions[otheroption_num], "");
	}

	// Back up one cuz that last space should be a CRLF
	buffer--;
	*buffer = '\n';

	return 0;
}

/**
 * @brief Encode a request in the Squid 3.0 protocol
 *
 * @param conf the current redirector configuration options
 * @param input the structure which holds the request data
 * @buffer the buffer in which the request should be created
 * @buffer_len the maximum number of bytes that can be put into the buffer
 *
 * @return 0 if the request was encoded successfully
 *        -1 if there was an error, and errno will be set:
 *           ENOMEM - There was a memory allocation error
 */
int squid30_encode (GlobalConf *conf, const protocol_input_line *input, char *buffer, size_t buffer_len)
{
	int otheroption_num;

	// Make sure we have all the data we need
	if (input->url == NULL)
	{
		errno = ENOENT;
		return -1;
	}

	// Start writing
	if( input->req_id != NULL && conf->server_concurrency) {
		BUFFER_WRITE_MACRO(buffer, buffer_len, input->req_id, "-");
	}
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->url, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->IP, "-/-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->username, "-");
	BUFFER_WRITE_MACRO(buffer, buffer_len, input->method, "-");
	// Squid 3.0 doesnt use the group, so don't write that

	// Now write our other options if we have them
	for (otheroption_num = 0;
		otheroption_num < MAX_PROTOCOLS_OTHER_OPTIONS &&
		  input->otheroptions[otheroption_num] != NULL;
		otheroption_num++)
	{
		BUFFER_WRITE_MACRO(buffer, buffer_len, input->otheroptions[otheroption_num], "");
	}

	// Back up one cuz that last space should be a CRLF
	buffer--;
	*buffer = '\n';

	return 0;
}

/**
 * @brief Encode a request in the Squid 3.1 protocol
 *
 * @param conf the current redirector configuration options
 * @param input the structure which holds the request data
 * @buffer the buffer in which the request should be created
 * @buffer_len the maximum number of bytes that can be put into the buffer
 *
 * @return 0 if the request was encoded successfully
 *        -1 if there was an error, and errno will be set:
 *           ENOMEM - There was a memory allocation error
 */
int squid31_encode (GlobalConf *conf, const protocol_input_line *input, char *buffer, size_t buffer_len)
{
	return squid30_encode(conf, input, buffer, buffer_len);
}

/**
 * @brief Encode a request in the Squid 3.2 protocol
 *
 * @param conf the current redirector configuration options
 * @param input the structure which holds the request data
 * @buffer the buffer in which the request should be created
 * @buffer_len the maximum number of bytes that can be put into the buffer
 *
 * @return 0 if the request was encoded successfully
 *        -1 if there was an error, and errno will be set:
 *           ENOMEM - There was a memory allocation error
 */
int squid32_encode (GlobalConf *conf, const protocol_input_line *input, char *buffer, size_t buffer_len)
{
	return squid30_encode(conf, input, buffer, buffer_len);
}

/********************************************************
 **************** Helper Function Definitions ***********
 *******************************************************/

/**
 * @brief Split a string containing space-separated fields into an array
 *
 * @param line the line we are splitting
 * @param fields the array into which the split fields should be put
 * @param max_fields the maximum number of fields that can be put in the array
 *
 * @return -1 if we can't fit all of the fields in the array
 *         otherwise, the number of fields we put in the array
 */
int protocol_explode_line (char *line, char **fields, size_t max_fields)
{
	char *strtok_save=NULL;
	size_t field_count = 0;
	char *field;

	field = strtok_r (line, " ", &strtok_save);

	// Parse line with strtok.  build up pointers
	//
	// strtok_r does not count an empty field as a field at all
	// so multiple spaces in a row or space on end of line will not
	// count as an additional field
	// Also empty lines count as zero fields not one field
	while (field_count < max_fields && field != NULL)
	{
		fields[field_count++] = field;
		field = strtok_r (NULL, " ", &strtok_save);
	}

	// If we have any additional fields then it's an error
	if (field != NULL && field_count == max_fields)
	{
		return -1;
	}
	return (int)field_count;
}

/**
 * @brief Check the first field in an array of fields to see if it's a request ID
 *
 * The first field is considered a request ID if its first character is a
 * decimal number.
 *
 * @fields the array of fields whose first member we should check
 * @field_cnt the number of fields in the array
 *
 * @return 0 (false) if the first field is not an ID
 *         1 (true) if the first field is an ID
 */
int protocol_def_squid_reqid (char **fields, size_t field_cnt)
{
	char *ptr;

	// Obviously no request id if field coutn is less than 1
	if (field_cnt < 1)
	{
		return 0;
	}

	// If the first field consists of digits
	// then that first field is a request id
	for (ptr = fields[0]; *ptr; ptr++) {
		if (!isdigit(*ptr)) {
			return 0;
		}
	}

	return -1;
}

/**
 * @brief Parse a line of Squid 3.0 redirector protocol input.
 *
 * @param input the container of the parsed data fields
 * @param fields the input data fields
 * @param field_cnt the number of fields remaining to be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int protocol_def_squid30_vparse (protocol_input_line *input, char **fields, size_t field_cnt)
{
	char *sptr;
	int otheroption_cnt;
	// if we have a request ID then pull it out and shift our fields
	// one to the right
	if (protocol_def_squid_reqid (fields, field_cnt) != 0)
	{
		input->req_id = strdup (fields[0]);
		fields++;
		field_cnt--;
	}

	// Now make sure we have at least 4 fields
	// (Since in squid 3.0, you don't actually have to have ANY
	//  key=value pairs)
	if (field_cnt < 4)
	{
		errno = EIO;
		return -1;
	}

	// The first 4 fields must be just like squid25
	if (protocol_def_squid25_vparse (input, fields, 4) != 0)
	{
		return -1;
	}

	// Squid 3.0 has no urlgroup in it

	// Loop through all the key value pairs and pass them along as
	// 'otheroption's
	otheroption_cnt = 0;
	// to make this easy, let's just shift our fields over by 4
	fields += 4;
	field_cnt -= 4;
	// Now keep looping while we still have fields.  Assume these
	// are extra options
	while (field_cnt > 0 && otheroption_cnt < MAX_PROTOCOLS_OTHER_OPTIONS)
	{
		// check if options are in format of key=value
		sptr = strchr (fields[0], '=');
		if (sptr == NULL || sptr == fields[0] || *(sptr + 1) == '\0')
		{
			errno = EIO;
			return -1;
		}
		input->otheroptions[otheroption_cnt] = strdup (fields[0]);
		if (input->otheroptions[otheroption_cnt] == NULL)
		{
			errno = ENOMEM;
			return -1;
		}
		otheroption_cnt++;
		field_cnt--;
		fields++;
	}
	// Did we get all the fields or did we hit our maximum option count?
	if (field_cnt > 0)
	{
		errno = EIO;
		return -1;
	}

	return 0;
}

/**
 * @brief Parse a line of Squid 2.7 redirector protocol input.
 *
 * @param input the container of the parsed data fields
 * @param fields the input data fields
 * @param field_cnt the number of fields remaining to be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int protocol_def_squid27_vparse (protocol_input_line *input, char **fields, size_t field_cnt)
{
	char *sptr;
	int otheroption_cnt;
	// if we have a request ID then pull it out and shift our fields
	// one to the right
	if (protocol_def_squid_reqid (fields, field_cnt) != 0)
	{
		input->req_id = strdup (fields[0]);
		fields++;
		field_cnt--;
	}

	// Now make sure we have at least 4 fields
	// (Since in squid 2.7, you don't actually have to have ANY
	//  key=value pairs)
	if (field_cnt < 5)
	{
		errno = EIO;
		return -1;
	}

	// The first 4 fields must be just like squid25
	if (protocol_def_squid25_vparse (input, fields, 4) != 0)
	{
		return -1;
	}

	// Squid 2.7 has urlgroup in it
	input->group = strdup (fields[4]);
	if (input->group == NULL)
	{
		errno = ENOMEM;
		return -1;
	}

	// Loop through all the key value pairs and pass them along as
	// 'otheroption's
	otheroption_cnt = 0;
	// to make this easy, let's just shift our fields over by 5
	fields += 5;
	field_cnt -= 5;
	// Now keep looping while we still have fields.  Assume these
	// are extra options
	while (field_cnt > 0 && otheroption_cnt < MAX_PROTOCOLS_OTHER_OPTIONS)
	{
		// check if options are in format of key=value
		sptr = strchr (fields[0], '=');
		if (sptr == NULL || sptr == fields[0] || *(sptr + 1) == '\0')
		{
			errno = EIO;
			return -1;
		}
		input->otheroptions[otheroption_cnt] = strdup (fields[0]);
		if (input->otheroptions[otheroption_cnt] == NULL)
		{
			errno = ENOMEM;
			return -1;
		}
		otheroption_cnt++;
		field_cnt--;
		fields++;
	}
	// Did we get all the fields or did we hit our maximum option count?
	if (field_cnt > 0)
	{
		errno = EIO;
		return -1;
	}

	return 0;
}

/**
 * @brief Parse a line of Squid 2.6 redirector protocol input.
 *
 * @param input the container of the parsed data fields
 * @param fields the input data fields
 * @param field_cnt the number of fields remaining to be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int protocol_def_squid26_vparse (protocol_input_line *input, char **fields, size_t field_cnt)
{
	// if we have a request ID then pull it out and shift our fields
	// one to the right
	if (protocol_def_squid_reqid (fields, field_cnt) != 0)
	{
		input->req_id = strdup (fields[0]);
		fields++;
		field_cnt--;
	}

	// Now make sure we have exactly 5 fields
	if (field_cnt != 5)
	{
		errno = EIO;
		return -1;
	}

	// The first 4 fields must be just like squid25
	if (protocol_def_squid25_vparse (input, fields, field_cnt - 1) != 0)
	{
		return -1;
	}

	// Finally, parse and validate the last field which in squid 2.6
	// is a group
	input->group = strdup (fields[4]);
	if (input->group == NULL)
	{
		errno = ENOMEM;
		return -1;
	}

	// There's no check we can run on the group.  So nothing more to validate
	// we're good
	return 0;
}

/**
 * @brief Parse a line of Squid 2.5 redirector protocol input.
 *
 * @param input the container of the parsed data fields
 * @param fields the input data fields
 * @param field_cnt the number of fields remaining to be parsed
 *
 * @return 0 if parsing was successful
 *        -1 if there was an error, and errno will be set:
 *           EIO - The input line was invalid
 *           ENOMEM - There was a memory allocation error
 */
int protocol_def_squid25_vparse (protocol_input_line *input, char **fields, size_t field_cnt)
{
	char *sptr;

	// We expect exactly 4 fields
	if (field_cnt != 4)
	{
		errno = EIO;
		return -1;
	}

	// Now just plug them in
	input->url = strdup (fields[0]);
	input->IP = strdup (fields[1]);
	input->username = strdup (fields[2]);
	input->method = strdup (fields[3]);

	if (input->url == NULL ||
		input->IP == NULL ||
		input->username == NULL ||
		input->method == NULL)
	{
		errno = ENOMEM;
		return -1;
	}

	// Finally, do some simple checks for validity of fields

	// If the method != CONNECT we need a scheme
	if ( strcmp("CONNECT", input->method) ) {
		// Every url must have a scheme (dont assume http)
		if (strstr (input->url, "://") == NULL)
		{
			errno = EIO;
			return -1;
		}
	}

	// The IP is expected to contain a forward slash with at least
	// one char on each side (ie -/-)
	sptr = strchr (input->IP, '/');
	if (sptr == NULL || sptr == input->IP || *(sptr + 1) == '\0')
	{
		errno = EIO;
		return -1;
	}

	return 0;
}
