/**
 * @file
 * @brief Test client username append domain to look like email address
 *
 * 

 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

// Arbitrarily-sized length for username we receive from redir
#define USERNAME_REWRITE_MAX_USERNAME_LEN 256

/**
 * Calling the QUOTE() macro will replace its argument (which must be
 * #define'd) with a quoted string of the value of that argument. Indirect
 * macro quoting expansion.
 *
 * They're used in the sscanf, where you have to specify the maximum field
 * width as part of the format string.
 */
#define _QUOTE(arg) #arg
#define QUOTE(arg) _QUOTE(arg)

/* @brief Info about the mock server.
 *
 * Static so init and cleanup can both get to it.
 */
static struct socketserver *svr1;

/* @brief Pointer to the redir process
 *
 * Static so init and cleanup can both get to it.
 */
static JFILE *fp;


/**
 * @brief Set up information about the socket connection
 *
 * This callback does nothing at present.
 */
static void connect_callback(__attribute__((unused)) struct client_socket* s)
{
	// No setup required
}

/**
 * @brief Allow the request in the socket rx buffer
 *
 * @param s the socket and associated buffers and statistics
 */
static void recv_allow_callback(struct client_socket* s)
{
	// Make sure the buffer has the whole request
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);

		// Consume the data
		s->rxlen = 0;
	}
}

/**
 * @brief Do cleanup before socket disconnect
 *
 * This callback does nothing at present.
 */
static void disconnect_callback(__attribute__((unused)) struct client_socket* s)
{
	// No cleanup required
}

/**
 * @brief Create the server and run redir
 *
 * This function sets up the environment for the tests by creating our mock
 * server and running the redir.
 *
 * @return 0 if initialization was successful
 *        -1 on error
 */
static int append_domain_initialize()
{
	// Create the server and always allow the requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);

	// Username rewriting MUST be enabled with the -u option
	fp = jopen("../redir -P12345 -i squid2.6 -o squid2.6 -H 192.168.4.0/24:office.com;10.10.0.0/16:other.com -L 3 localhost");

	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return -1;
	}

	return 0;
}

/**
 * @brief Clean up the test environment
 *
 * This function cleans up the environment by shutting down the server and
 * stopping the redir process.
 *
 * @return 0 if initialization was successful
 *        -1 on error
 */
static int append_domain_cleanup()
{
	jclose(fp);

	socketserver_signal_shutdown(svr1);
	socketserver_waitfor_shutdown(svr1);

	return 0;
}

/**
 * @brief Send a request to the server with a specific username
 *
 * @param input_username the username to send to the redir
 * @param expected_username the username we expect the redir to send to the
 * policy server
 *
 * @return 0 if we got the username we expected
 *        -1 if we failed (i.e. didn't get the username we expected)
 */
static int append_domain_send_request(const char *input_username,
					 const char *ip,
					 const char *expected_username)
{
	char sendbuffer[256];
	char buffer[256];
	char received_output[USERNAME_REWRITE_MAX_USERNAME_LEN];
	int ret;

	// Create the request that includes the input username
	sprintf(sendbuffer, "1234 http://google.ca/ %s/- %s - -\n", ip, input_username);

	// Send the request to the redir
	ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

	// Get the request from the redir to the policy server
	ret=read(fp->from_stdout, buffer, sizeof(buffer));
	if ( ret >= 0 ) {
		buffer[ret]='\0';
	}

	if ( ret < 0 ) {
		fprintf (stderr, "error reading from stdout\n");
		return -1;
	}


	// Get only the username from the request, and limit its size
	if ( sscanf(buffer, "%*s %*s %*s %" QUOTE(USERNAME_REWRITE_MAX_USERNAME_LEN) "s %*s %*s", received_output) == 0 ) {
		// If we failed to convert in sscanf, terminate the username
		received_output[0] = '\0';
	}
	

	// Error if the output does not match what's expected
	if ( strcmp(received_output, expected_username) != 0) {
		//fprintf(stderr, "Failed: expected '%s', received '%s'\n", expected_username, received_output);
		return -1;
	}

	return 0;
}


// These tests verify that valid characters are not replaced

static void office_user_test()
{
	CU_ASSERT(append_domain_send_request("admin", "192.168.4.100", "admin@office.com") == 0);
}

static void other_user_test()
{
	CU_ASSERT(append_domain_send_request("admin", "10.10.0.1", "admin@other.com") == 0);
}

static void outside_user_test()
{
	CU_ASSERT(append_domain_send_request("admin", "172.31.2.1", "admin") == 0);
}

static void noip_user_test()
{
	CU_ASSERT(append_domain_send_request("admin", "-", "admin") == 0);
}




/**
 * @brief Register our test cases and test suite with CUnit
 *
 * @return 0 if we registered everything successfully
 *         nonzero otherwise (returns CU_ErrorCode errors)
 */
int append_domain_test_init()
{
	// Username rewrite tests
	CU_TestInfo append_domain_tests[] = {
		// Don't replace valid characters tests
		{ "User in first subnnet",      	  office_user_test }, 
		{ "User in second subnet",                other_user_test }, 
		{ "User from outside IP",                 outside_user_test }, 
		{ "User and no IP",                       noip_user_test }, 

		CU_TEST_INFO_NULL
	};

	// List our test suites, which contain tests
	CU_SuiteInfo append_domain_suites[] = {
		{ "Append Domain Test Suite", append_domain_initialize, append_domain_cleanup, append_domain_tests },
		CU_SUITE_INFO_NULL
	};
	
	// Register our test suites with CUnit
	if (CU_register_suites(append_domain_suites) != CUE_SUCCESS) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
