/**
 * @file
 * @brief Utils File which contains the logic for GlobalCounters and Queue processing
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */


#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <pthread.h>
#include <syslog.h>
#include <time.h>
#include <errno.h>

#include "redir.h"
#include "protocol.h"
#include "utils.h"
#include "request.h"
#include "config.h"
#include "log.h"

/** @brief Global counter for the number of servers available and connected */
static GlobalCount servers_available = GLOBAL_COUNT_INIT;
/** @brief Global counter for the number of threads (connections) we have in total */
static GlobalCount running_threads = GLOBAL_COUNT_INIT;

/** @brief Our global queue of requests */
static ReqQueue* global_queue = NULL;
/** @brief The maximum number of requests */
static int max_requests = 1;
/** @brief A pointer to the global configuration */
static GlobalConf* global_config = NULL;

/** @brief Condition we use when our queue if full */
static pthread_cond_t queue_full = PTHREAD_COND_INITIALIZER;
/** @brief Lock used for condition when queue is full */
static pthread_mutex_t queue_full_lock = PTHREAD_MUTEX_INITIALIZER;

/** @brief Condition we use when we are going to flush request queue */
static pthread_cond_t need_flush_queue = PTHREAD_COND_INITIALIZER;
/** @brief Lock used for condition when we are going to flush the request queue */
static pthread_mutex_t need_flush_queue_lock = PTHREAD_MUTEX_INITIALIZER;
/** @brief The flag that shows that we are going to flush queue after a timeout */
static int flush_pending = 0;
/** @brief The absolute time when queue flushing can really start */
static struct timeval no_flush_until;
/** @brief The ID of the thread that does queue flushing */
static pthread_t flushing_thread_id;

// Manage the total count of server threads that are able to serv requests
// if no one is available the redir shoul not queue requests

/* @brief Increment the active connections counter
 */

void redir_st_up()
{
	// increase the number of available server connections
	pthread_mutex_lock(&servers_available.lock);
	servers_available.num++;
	pthread_mutex_unlock(&servers_available.lock);
}

/* @brief Decrement the active connections counter
 *
 * It also starts delayed flushing for the request queue
 */
void redir_st_down()
{
	// decrease the number of available server connections
	pthread_mutex_lock(&servers_available.lock);
	servers_available.num--;

	// if all servers are down flush the request queue
	if (servers_available.num == 0) {
		// wake up the flushing thread
		redir_start_flush_request_queue();
	}

	pthread_mutex_unlock(&servers_available.lock);
}

/* @brief Retun the current number of active connections
 */
int redir_get_servers()
{
	int ret;

	pthread_mutex_lock(&servers_available.lock);
	ret = servers_available.num;
	pthread_mutex_unlock(&servers_available.lock);

	return ret;
}

// Manage the total count of running server threads
// if no one is running the redir should exit

/* @brief Increment the running threads counter
 */
void redir_st_run()
{
	// increase the number of available server connections
	pthread_mutex_lock(&running_threads.lock);
	running_threads.num++;
	pthread_mutex_unlock(&running_threads.lock);
}

/* @brief Decrement the running threads counter
 */
void redir_st_exit()
{
	// decrease the number of available server connections
	pthread_mutex_lock(&running_threads.lock);
	running_threads.num--;
	pthread_mutex_unlock(&running_threads.lock);
}

/* @brief Return the current number of running threads
 */
int redir_get_threads()
{
	int ret;

	pthread_mutex_lock(&running_threads.lock);
	ret = running_threads.num;
	pthread_mutex_unlock(&running_threads.lock);

	return ret;
}

// Manage the requests queue
// It is a simple linked list with phread_cond synchronization

/* @brief Allocate and init the queue handler
 *
 * return allocated queue handler
 */
ReqQueue* redir_init_queue()
{
	// allocate the header
	ReqQueue* queue = (ReqQueue *)xmalloc(sizeof(ReqQueue));

	// init pointers and counter
	queue->head = queue->tail = NULL;
	queue->size = 0;

	// init pthread mutex and condition
	pthread_mutex_init(&queue->lock, NULL);
	pthread_cond_init(&queue->cond, NULL);

	return queue;
}

/* @brief Free the queue handler
 *
 * @param queue The queue handler
 */
void redir_free_queue(ReqQueue* queue)
{
	if (queue == NULL) {
		return;
	}

	pthread_mutex_destroy(&queue->lock);
	pthread_cond_destroy(&queue->cond);

	free(queue);
}

/* @brief Append a request to the queue
 *
 * It appends a request to the list and wakes up
 * a thread that waits request
 *
 * @param queue The queue handler
 * @param req The request to append
 */
void redir_queue_push (ReqQueue* queue, void* req)
{
	// lock the queue
	pthread_mutex_lock(&queue->lock);

	((Request*)req)->next = NULL;

	// set list pointers
	if (queue->head == NULL) {
		queue->head = req;
		queue->tail = req;
	} else {
		queue->tail->next = req;
		queue->tail = req;
	}
	// count the request
	queue->size++;

	// wake up a thread waiting request
	pthread_cond_signal(&queue->cond);
	// unlock the queue
	pthread_mutex_unlock(&queue->lock);
}

/* @brief Insert a request to the queue head
 *
 * It inserts a request into the list and wakes up
 * a thread that waits request
 *
 * @param queue The queue handler
 * @param req The request to append
 */
void redir_queue_unshift (ReqQueue* queue, void* req)
{
	// lock the queue
	pthread_mutex_lock(&queue->lock);

	((Request*)req)->next = queue->head;

	// set list pointers
	if (queue->head == NULL) {
		queue->head = req;
		queue->tail = req;
	} else {
		queue->head = req;
	}
	// count the request
	queue->size++;

	// wake up a thread waiting request
	pthread_cond_signal(&queue->cond);
	// unlock the queue
	pthread_mutex_unlock(&queue->lock);
}

/* @brief Get a request from the queue
 *
 * If the queue is empty the procedure waits
 * until some data is added to the queue
 *
 * @param queue The queue handler
 * return The request to append
 */
Request* redir_queue_pop (ReqQueue* queue, Connection* conn, int timeout)
{
	Request* ret;

	// lock the queue
	pthread_mutex_lock(&queue->lock);

	// if there is nothig to return
	if (queue->size == 0) {
		struct timeval tv;
		struct timespec ts;

		gettimeofday(&tv, NULL);
		ts.tv_sec = tv.tv_sec + timeout / 1000;
		ts.tv_nsec = tv.tv_usec*1000 + (timeout % 1000)*1000000;
		if (ts.tv_nsec >= 1000000000) {
			ts.tv_nsec -= 1000000000;
			ts.tv_sec++;
		}

		// wait new requests
		pthread_cond_timedwait(&queue->cond, &queue->lock, &ts);

		// make sure the connection is not dead,
		// too complicated.
		// Go ahead to precess it, the server thread will figure out it can't be sent.
		// then throw it back to the queue. Keep the error handling in one place.
		if (st_check_connection(conn) < 0) {
			pthread_cond_signal(&queue->cond);
			pthread_mutex_unlock(&queue->lock);
			return NULL;
		}

                if (queue->size == 0) {
			// return to caller
			pthread_mutex_unlock(&queue->lock);
			return NULL;
		}
	}

	// set list pointers
	ret = queue->head;
	queue->head = ret->next;

	if (queue->head == NULL) {
		queue->tail = NULL;
	}

	// decrement the number of queued elements
	queue->size--;

	// unlock the queue
	pthread_mutex_unlock(&queue->lock);

	return ret;
}

/* @brief Get a request from the queue without waiting
 *
 * If the queue is empty the procedure returns NULL
 * It is used for flushing teh queue
 *
 * @param queue The queue handler
 * return The request to append
 */

Request* redir_queue_walk (ReqQueue* queue)
{
	Request* ret;

	// lock the queue
	pthread_mutex_lock(&queue->lock);

	// if the queue is empty
	if (queue->size == 0) {
		// unlock the queue and return "nothing"
		pthread_mutex_unlock(&queue->lock);
		return NULL;
	}
	// set list pointers
	ret = queue->head;
	queue->head = ret->next;

	if (queue->head == NULL) {
		queue->tail = NULL;
	}

	// decrement the number of queued elements
	queue->size--;
	// unlock the queue
	pthread_mutex_unlock(&queue->lock);

	return ret;
}

/* @brief Returns the number of elements in the queue
 *
 * return The number of elements
 */

int redir_queue_size(ReqQueue* queue)
{
	return queue->size;
}

// Dispatcher API
// Now it uses one global queue for all threads

/* @brief Init all data for dispatcher
 *
 * return 0 - success, -1 -error
 */
int redir_dispatcher_init(GlobalConf *conf)
{
	// init the global queue (static variable)
	global_queue = redir_init_queue();

	// check if the queue handler was really allocated
	if (global_queue == NULL) {
		log_send(LOG_ERR, "Can not allocate memory");
		return -1;
	}

	// fill static data
	max_requests = conf->max_requests;
	global_config = conf;

	if (pthread_create(&flushing_thread_id, NULL, queue_flush_thread, NULL) != 0) {
		log_send(LOG_ERR, "Can not start queue maintenance thread\n");
		return -1;
	}

	return 0;
}

void redir_dispatcher_terminate()
{
	Request *ptr;
	void *exit_data;

	// terminate flushing queue
	pthread_mutex_lock(&need_flush_queue_lock);
	pthread_cond_signal(&need_flush_queue);
	pthread_mutex_unlock(&need_flush_queue_lock);

	pthread_join(flushing_thread_id, &exit_data);

	// discard all remaining requests
	while ((ptr = redir_queue_walk(global_queue)) != NULL) {
		redir_free_request(ptr);
	};

	redir_free_queue(global_queue);
}

/* @brief Dispatch a request for processing
 *
 * @param request the request to process
 */
void redir_dispatch_request(Request *request, int tohead)
{
	if ( (request->ttl)-- > 0 &&				// check if request retry count is not exceeded
	    (redir_get_servers() > 0 || flush_pending)) {	// check if we have server threads that can serve it
		// put the request to the global queue
		if (tohead) {
			redir_queue_unshift(global_queue, request);
		} else {
			redir_queue_push(global_queue, request);
		}
	} else {
		// if one of conditiion is not true don't queue the request
		// return the 'fail' reply
		redir_write_reply_failed(request, global_config);
	}
}

/* @brief Checks if dispatch can queue more requests
 *
 * It comes to sleep if queue(s) is full until at least
 * one request is removed from queue
 */
void redir_check_dispatch_space()
{
	int ret;
	struct timeval tv;
	struct timespec ts;

	// lock access to queue size condition
	pthread_mutex_lock(&queue_full_lock);

	while (!redir_should_exit) {
		// check for a free space in the queue
		if ( redir_queue_size(global_queue) < max_requests) {
			break;
		}
		// wait queue size decriment
		do {
			gettimeofday(&tv, NULL);
			ts.tv_sec = tv.tv_sec + 1;
			ts.tv_nsec = tv.tv_usec * 1000;

			ret = pthread_cond_timedwait(&queue_full, &queue_full_lock, &ts);
		} while (ret == ETIMEDOUT && !redir_should_exit);
	}
	// unlock access to queue size condition
	pthread_mutex_unlock(&queue_full_lock);
}

/* @brief Get dispatched request for processing
 *
 * @param conn The processing thread (connection) handler
 */
Request* redir_pop_request(Connection *conn)
{
	Request *ret;

	// get a request from the global queue
	ret = redir_queue_pop(global_queue, conn, conn->global_conf->queue_wait_timeout);

	// lock access to queue size condition
	pthread_mutex_lock(&queue_full_lock);
	// wake up a thread that can wait
	pthread_cond_broadcast(&queue_full);
	// unlock access to queue size condition
	pthread_mutex_unlock(&queue_full_lock);

	return ret;
}

/**
 * @brief Thread function that flushes the queue of requests
 *
 * @param arg unusued
 *
 * @return NULL
 */
void * queue_flush_thread(void* arg __attribute__ ((unused)))
{
	struct timeval tv;

	// loop forever
	while (!redir_should_exit) {
		// sleep until the queue needs flushing
		pthread_mutex_lock(&need_flush_queue_lock);
		pthread_cond_wait(&need_flush_queue, &need_flush_queue_lock);
		pthread_mutex_unlock(&need_flush_queue_lock);

		// wait a while before really flush the queue
		// to allow servers threads to reconnect
		while (! redir_should_exit) {
			// if the timeout was set again when we was waiting
			// we have to wait again
			gettimeofday(&tv, NULL);

			if ( no_flush_until.tv_sec > tv.tv_sec ||
			    (no_flush_until.tv_sec == tv.tv_sec && no_flush_until.tv_usec > tv.tv_usec )) {

				tv.tv_sec = no_flush_until.tv_sec - tv.tv_sec;
				tv.tv_usec = no_flush_until.tv_usec - tv.tv_usec;
				if ( tv.tv_usec < 0 ) {
					tv.tv_usec += 1000000;
					tv.tv_sec--;
				}
				select( 0, NULL, NULL, NULL, &tv);
			} else {
				break;
			}
		}
		// check if we wake up just for exit
		if (redir_should_exit) {
			break;
		}

		// all timeouts expired we don't postpone flushing more
		flush_pending = 0;

		// if some threads become active during the delay
		// we don't need to flush the queue
		if (redir_get_servers()) {
			continue;
		}
		// otherwise really flush the queue
		redir_do_flush_request_queue(0);

		// wake up the main thread if it is waiting queue space
		// lock access to queue size condition
		pthread_mutex_lock(&queue_full_lock);
		// wake up a thread that can wait
		pthread_cond_broadcast(&queue_full);
		// unlock access to queue size condition
		pthread_mutex_unlock(&queue_full_lock);
	}

	// flush the queue
	redir_do_flush_request_queue(1);

	// wake up the main thread if it is waiting queue space
	// lock access to queue size condition
	pthread_mutex_lock(&queue_full_lock);
	// wake up a thread that can wait
	pthread_cond_broadcast(&queue_full);
	// unlock access to queue size condition
	pthread_mutex_unlock(&queue_full_lock);

	return NULL;
}

/* @brief Wakes up the thread that flushes queue
 */
void redir_start_flush_request_queue()
{
	struct timeval tv;

	// lock access to queue flush queue condition
	pthread_mutex_lock(&need_flush_queue_lock);

	// set the flag to prevent dispatcher from request rejecting
	// until timeout expired
	flush_pending = 1;

	// calculate the absolute time when flushing should really start
	gettimeofday(&tv, NULL);

	tv.tv_sec  += global_config->queue_flush_delay / 1000;
	tv.tv_usec += (global_config->queue_flush_delay % 1000) * 1000;
	if (tv.tv_usec >= 1000000) {
		tv.tv_usec -= 1000000;
		tv.tv_sec++;
	}
	no_flush_until = tv;

	// wake up a thread that will flush the queue after timeout
	pthread_cond_broadcast(&need_flush_queue);
	// unlock access to queue flush queue condition
	pthread_mutex_unlock(&need_flush_queue_lock);
}

void redir_do_flush_request_queue(int force)
{
	Request* ptr;

	// get all queue elements one by one
	while ((ptr = redir_queue_walk(global_queue)) != NULL) {
		// return the 'fail' reply
		redir_write_reply_failed(ptr, global_config);

		// we don't need to empty the queue
		// if some connection became active
		// (if 'force' flag is set flush anyway)
		if (!force && redir_get_servers()) {
			return;
		}
	}
}

void redir_dispatcher_broadcast()
{
	// lock the queue
	pthread_mutex_lock(&global_queue->lock);
	// wake up a thread waiting request
	pthread_cond_broadcast(&global_queue->cond);
	// unlock the queue
	pthread_mutex_unlock(&global_queue->lock);
}
